#include <Arduino.h>
#include <Adafruit_WS2801.h>
#include <Tone32.h>
#include <Adafruit_NeoPixel.h>

/**
 * Player 1: x = 0
 * Player 2: x = X_MAX-1
 */

#define PIN_LED_MATRIX        2
#define PIN_JOYSTICK_1_BUTTON 12
#define PIN_JOYSTICK_2_BUTTON 26
#define PIN_PIEZO             17
#define CHANNEL_PIEZO         1 
#define PIN_JOYSTICK_1_Y      13
#define PIN_JOYSTICK_2_Y      27
#define PIN_JOYSTICK_1_X      14
#define PIN_JOYSTICK_2_X      25

#define JOYSTICK_OFFSET_MIN 200
#define JOYSTICK_OFFSET_MAX 3500

#define DEBOUNCE_TIME 300  // in ms

#define X_MAX 10
#define Y_MAX 20

#define GAME_DELAY 40         // in ms
#define BALL_DELAY_MAX   200  // in ms
#define BALL_DELAY_MIN    50  // in ms
#define BALL_DELAY_STEP    10  // in ms

#define PLAYER_AMOUNT 2
#define PLAYER_1 0
#define PLAYER_2 1

#define PADDLE_WIDTH 3

#define PADDLE_MOVE_NONE  0
#define PADDLE_MOVE_UP    1
#define PADDLE_MOVE_DOWN  2

#define LED_TYPE_OFF           1
#define LED_TYPE_PADDLE_BLUE   2
#define LED_TYPE_PADDLE_RED    3
#define LED_TYPE_BALL          4
#define LED_TYPE_BALL_YELLOW   5
#define LED_TYPE_POINTS_BLUE   6
#define LED_TYPE_POINTS_RED    7 

#define TONE_PLAYER 1
#define TONE_WALL   2
#define TONE_BUZZ   3

#define GAME_STATE_RUNNING      1
#define GAME_STATE_END          2
#define GAME_STATE_INIT         3
#define GAME_STATE_INACTIVE     4
#define GAME_STATE_PLAY_WITH_ME 5

#define LIGHT_STATE_BAGGROUND  1
#define LIGHT_STATE_PLAY       2   
#define LIGHT_STATE_WITH       3
#define LIGHT_STATE_ME         4 
#define LIGHT_STATE_END        5


uint8_t dataPin  = 2;    
uint8_t clockPin = 4;

struct Coords {
  byte x;
  byte y;
};

byte pointsPlayerRed =0;
byte pointsPlayerBlue =0; 

Adafruit_WS2801 pixels = Adafruit_WS2801((uint16_t)10 ,(uint16_t)20, dataPin, clockPin );
bool buttonPressed = false;
byte gameState;
byte joystickPins[PLAYER_AMOUNT] = {PIN_JOYSTICK_1_X, PIN_JOYSTICK_2_X};
Coords paddles[PLAYER_AMOUNT][PADDLE_WIDTH];
Coords ball;
int ballMovement[2];
unsigned int ballDelay;
unsigned long lastDrawUpdate = 0;
unsigned long lastBallUpdate = 0;
unsigned long lastButtonClick = 0;

void resetLEDs();
void initGame();
void updateBall();
void updateGame();
boolean isButtonPressed();
void playTone(byte type);
void toggleLed(byte x, byte y, byte type);
byte getPlayerMovement(byte playerId);
void endGame();
void attentionLight(); 
void attentionLight1(); 
unsigned long delay_light = millis(); 
byte lightState = LIGHT_STATE_BAGGROUND;
boolean delay_light_high = true;  


static uint32_t Color(uint8_t r, uint8_t g, uint8_t b) {
    return ((uint32_t)r << 16) | ((uint32_t)g << 8) | b;
  }
  /*!
    @brief   Convert separate red, green, blue and white values into a
             single "packed" 32-bit WRGB color.
    @param   r  Red brightness, 0 to 255.
    @param   g  Green brightness, 0 to 255.
    @param   b  Blue brightness, 0 to 255.
    @param   w  White brightness, 0 to 255.
    @return  32-bit packed WRGB value, which can then be assigned to a
             variable for later use or passed to the setPixelColor()
             function. Packed WRGB format is predictable, regardless of
             LED strand color order.
  */
  static uint32_t Color(uint8_t r, uint8_t g, uint8_t b, uint8_t w) {
    return ((uint32_t)w << 24) | ((uint32_t)r << 16) | ((uint32_t)g << 8) | b;
  }
  static uint32_t ColorHSV(uint16_t hue, uint8_t sat = 255, uint8_t val = 255);
  /*!
    @brief   A gamma-correction function for 32-bit packed RGB or WRGB
             colors. Makes color transitions appear more perceptially
             correct.
    @param   x  32-bit packed RGB or WRGB color.
    @return  Gamma-adjusted packed color, can then be passed in one of the
             setPixelColor() functions. Like gamma8(), this uses a fixed
             gamma correction exponent of 2.6, which seems reasonably okay
             for average NeoPixels in average tasks. If you need finer
             control you'll need to provide your own gamma-correction
             function instead.
  */


void setup()
{
  ledcAttachPin(PIN_PIEZO , CHANNEL_PIEZO);
  ledcSetup(CHANNEL_PIEZO, 10000, 16);
  pinMode(PIN_JOYSTICK_1_Y, INPUT);
  pinMode(PIN_JOYSTICK_1_BUTTON, INPUT_PULLUP);
  pinMode(PIN_JOYSTICK_2_Y, INPUT);
  pinMode(PIN_JOYSTICK_2_BUTTON, INPUT_PULLUP);
  pixels.begin();
  resetLEDs();
  
  Serial.begin(9600);
  
  gameState = GAME_STATE_END;
 
}

void loop()
{
 
  switch(gameState) {
    case GAME_STATE_INIT:

      lightState = LIGHT_STATE_BAGGROUND; 
      delay_light_high = true; 
      initGame();
      break;
    case GAME_STATE_RUNNING:
      updateBall();
      updateGame();
      break;
    case GAME_STATE_END:
      if (isButtonPressed()) {
        gameState = GAME_STATE_INIT;
      }
      else if (millis() - lastDrawUpdate > 3000 && millis() - lastDrawUpdate < 30000 && millis() - delay_light >= 500 && false == delay_light_high || millis() - lastDrawUpdate > 15000 && millis() - lastDrawUpdate < 30000 && millis() - delay_light >= 500 && true == delay_light_high )
      {
        gameState = GAME_STATE_INACTIVE; 
      }

      else if (millis() - lastDrawUpdate > 30000 && millis() - delay_light >= 1500)
      {
        gameState = GAME_STATE_PLAY_WITH_ME; 
      }

      break;
    case GAME_STATE_INACTIVE: 

      resetLEDs(); 
      pointsPlayerBlue = 0; 
      pointsPlayerRed = 0;
      attentionLight1(); 
      gameState = GAME_STATE_END; 
    break;

    case GAME_STATE_PLAY_WITH_ME:

      resetLEDs(); 
      pointsPlayerBlue = 0; 
      pointsPlayerRed = 0; 
      attentionLight();
      gameState = GAME_STATE_END; 
    break;
  }
}

void initGame()
{
  
  resetLEDs();
  lastButtonClick = millis();

  ball.y = random(7,13);
  ball.x = random(3,6);
  switch(random(3))
  {
  case 0:
  
    ballMovement[0] =  -1;
    ballMovement[1] = 1;
    break;
  
  case 1: 
  
    ballMovement[0] =  1;
    ballMovement[1] = -1;
    break;

    case 2:
    ballMovement[0] = -1;
    ballMovement[1] = -1;
    break;

    case 3:
    ballMovement[0] =  1;
    ballMovement[1] =  1;
    break;
  }
  ballDelay = BALL_DELAY_MAX;

  for(byte i=0; i<PADDLE_WIDTH; i++) {
    paddles[PLAYER_1][i].y = 0;
    paddles[PLAYER_1][i].x = (X_MAX/2) - (PADDLE_WIDTH/2) + i;
    paddles[PLAYER_2][i].y = Y_MAX - 1;
    paddles[PLAYER_2][i].x = paddles[PLAYER_1][i].x;
  }

  gameState = GAME_STATE_RUNNING;
}

void updateBall()
{
  bool hitBall = false;
  if ((millis() - lastBallUpdate) < ballDelay) {
    return;
  }
  lastBallUpdate = millis();
  toggleLed(ball.x, ball.y, LED_TYPE_OFF);

  Serial.printf("ballMovement[1]=%d, ball.y=%d, ball.x=%d\n", ballMovement[1], ball.y, ball.x);

  byte next_ball_x = ball.x + ballMovement[0];
  byte next_ball_y = ball.y + ballMovement[1];

  // collision detection for player 1
  if (ballMovement[1] == -1 && next_ball_y == 0) {
    Serial.println("Check Player 1 ...");
    for(byte i=0; i<PADDLE_WIDTH; i++) {
      Serial.printf("x[%d]=%d\n", i, paddles[PLAYER_1][i].x );
      if (paddles[PLAYER_1][i].x == next_ball_x) {
        Serial.println("HIT!! Player 1");
        hitBall = true;
        
        break;
      }
       
    }
    if(hitBall == false)
    {
      pointsPlayerRed++;
    }
    
  }

  // collision detection for player 2
  if (ballMovement[1] == 1 && next_ball_y == Y_MAX-1) {
    Serial.println("Check Player 2 ...");
    for(byte i=0; i<PADDLE_WIDTH; i++) {
      Serial.printf("x[%d]=%d\n", i, paddles[PLAYER_1][i].x );
      if (paddles[PLAYER_2][i].x == next_ball_x) {
        Serial.println("HIT!! Player 1");
        hitBall = true;
        
        break;
      }
      
    }
    if (hitBall == false)
    {
      pointsPlayerBlue++;
    }
     
  }

  if (hitBall == true) {
    ballMovement[1] *= -1;
    playTone(TONE_PLAYER);
    if (ballDelay > BALL_DELAY_MIN) {
      ballDelay -= BALL_DELAY_STEP;
    }
  }
  else{
    if (ball.y <=0 || ball.y >= Y_MAX-1) {
      endGame();
      return;
    }
  }

  ball.x += ballMovement[0];
  ball.y += ballMovement[1];
  

  if (ball.x <= 0 || ball.x >= X_MAX-1) {
    ballMovement[0] *= -1;
    playTone(TONE_WALL);
  }

  toggleLed(ball.x, ball.y, LED_TYPE_BALL);
  pixels.show();
}

void endGame()
{
  gameState = GAME_STATE_END;
  toggleLed(ball.x, ball.y, LED_TYPE_BALL_YELLOW);
  pixels.show();
  playTone(TONE_BUZZ);
}

void updateGame()
{
  if ((millis() - lastDrawUpdate) < GAME_DELAY) {
    return;
  }
  lastDrawUpdate = millis();

  // turn off paddle LEDs
  for(byte p=0; p<PLAYER_AMOUNT; p++) {
    for(byte i=0; i<PADDLE_WIDTH; i++) {
      toggleLed(paddles[p][i].x, paddles[p][i].y, LED_TYPE_OFF);
    }
  }

  // move paddles
  for(byte p=0; p<PLAYER_AMOUNT; p++) {
    byte movement = getPlayerMovement(p);
    if (movement == PADDLE_MOVE_UP && paddles[p][PADDLE_WIDTH-1].x < (X_MAX-1)) {
      for(byte i=0; i<PADDLE_WIDTH; i++) {
        paddles[p][i].x++;
      }
    }
    if (movement == PADDLE_MOVE_DOWN && paddles[p][0].x > 0) {
      for(byte i=0; i<PADDLE_WIDTH; i++) {
        paddles[p][i].x--;
      }
    }
  }

  // show paddle LEDs
  for(byte p=0; p<PLAYER_AMOUNT; p++) {
    for(byte i=0; i<PADDLE_WIDTH; i++) {
      toggleLed(paddles[p][i].x, paddles[p][i].y, LED_TYPE_PADDLE_BLUE+p);
    }
  }

  

  // show points 
  for(int i = 0; i<pointsPlayerBlue; i++)
  {
    Serial.printf("Punkte Player Blue %d\n", pointsPlayerBlue);

    toggleLed(0,9-i, LED_TYPE_POINTS_BLUE);
  }
  for(int i = 0; i<pointsPlayerRed; i++)
  {
    Serial.printf("Punkte Player Red %d\n", pointsPlayerRed);
    toggleLed(0,10+i,LED_TYPE_POINTS_RED);
  }
      if(pointsPlayerBlue == 5)
    {

      pointsPlayerBlue = 0; 
      pointsPlayerRed = 0; 
      for (int i =0; i<200; i++)
      {
        pixels.setPixelColor(i, 0,0,55);
      }
      
    }
    else if(pointsPlayerRed == 5)
    {
      pointsPlayerBlue = 0; 
      pointsPlayerRed = 0;
      for (int i =0; i<200; i++)
      {
        pixels.setPixelColor(i, 55,0,0);
      }
      
      
    }
  pixels.show();
}

byte getPlayerMovement(byte playerId)
{
  int value = analogRead(joystickPins[playerId]);
  if (value < JOYSTICK_OFFSET_MIN) {
    //if (playerId == PLAYER_2) {
      return PADDLE_MOVE_DOWN;
    //} else {
    //  return PADDLE_MOVE_UP;
    //}
  } else if (value > JOYSTICK_OFFSET_MAX) {
    //if (playerId == PLAYER_2) {
      return PADDLE_MOVE_UP;
    //} else {
    //  return PADDLE_MOVE_DOWN;
    //}
  }
  return PADDLE_MOVE_NONE;
}

bool isButtonPressed()
{
  if ((millis() - lastButtonClick) < DEBOUNCE_TIME) {
    return false;
  }
  if (digitalRead(PIN_JOYSTICK_1_BUTTON) == LOW) {
    lastButtonClick = millis();
    return true;
  }
  if (digitalRead(PIN_JOYSTICK_2_BUTTON) == LOW) {
    lastButtonClick = millis();
    return true;
  }
  return false;
}

void resetLEDs()
{
  for(byte i=0; i<X_MAX*Y_MAX; i++) {
    pixels.setPixelColor(i, 0, 0, 0);
  }
  pixels.show();
}

void toggleLed(byte x, byte y, byte type)
{
  byte ledIndex ;
  if(y%2 == 0)
  {
    ledIndex = y * X_MAX + x; 
  }
  else
  {
    ledIndex = y * X_MAX + (X_MAX -x -1); 
  }
  uint32_t color;

  switch(type) {
    case LED_TYPE_PADDLE_RED:
      color = Color(255, 0, 0);
      break;
    case LED_TYPE_PADDLE_BLUE:
      color = Color(0,0,255);
      break; 
    case LED_TYPE_BALL_YELLOW:
      color = Color(255, 255, 0);
      break;
    case LED_TYPE_BALL:
      color = Color(0, 255, 0);
      break;
    case LED_TYPE_OFF:
      color = Color(0, 0, 0);
      break;
    case LED_TYPE_POINTS_RED:
      color = Color(64,0,0);
      break; 
    case LED_TYPE_POINTS_BLUE:
      color = Color(0,0,64);
      break; 
  }

  pixels.setPixelColor(ledIndex, color);
}

void playTone(byte type)
{
  switch(type) {
    case TONE_PLAYER:
      tone(PIN_PIEZO, NOTE_A4, 50,0);
      break;
    case TONE_WALL:
      tone(PIN_PIEZO, NOTE_CS5, 50,0);
      break;
    case TONE_BUZZ:
      for(byte i=0; i<20; i++) {
        tone(PIN_PIEZO, NOTE_A3-i*10, 50,0);
        delay(50);
      }
      break;
  }
}


void attentionLight()
{
  int play [44] = {18,17,16,15,14,13,26,33,34,35,24,53,54,55,56,57,58,61,78,81,102,103,104,105,106,101,113,126,133,134,135,136,137,138,124,115,178,177,176,175,165,153,185,193};
  int with [46] = {13,14,15,16,17,18,22,36,42,58,57,56,55,54,53,73,74,75,76,77,78,93,106,114,126,133,115,116,117,118,113,154,155,156,157,158,153,163,176,183,193,195,196,197,198,194};
  int me   [30]   = {53,54,55,56,57,58,65,75,85,93,94,95,96,97,98,113,114,115,116,117,118,126,138,146,121,133,144,124,135,141};

  switch (lightState)
  {
  case LIGHT_STATE_BAGGROUND:

   for (int i = 0; i< 200; i++)
    {
      pixels.setPixelColor(i,Color(0,5,5));       
    }

    pixels.show(); 
    lightState = LIGHT_STATE_PLAY;
    delay_light = millis();
    
  break;

  case LIGHT_STATE_PLAY:

    for (int i = 0; i<44; i++)
      {
        pixels.setPixelColor(play[i],Color(255,0,0));       
      }

    pixels.show(); 
    lightState = LIGHT_STATE_WITH; 
    delay_light = millis(); 
  break; 

  case LIGHT_STATE_WITH: 

   for (int i = 0; i< 200; i++)
    {
      pixels.setPixelColor(i,Color(5,5,5));
    }
  

   for (int i = 0; i<46; i++)
    {
      pixels.setPixelColor(with[i],Color(255,0,0));
    }

    pixels.show();
    lightState = LIGHT_STATE_ME; 
    delay_light = millis(); 
  break;

  case LIGHT_STATE_ME:

    for (int i = 0; i< 200; i++)
      {
        pixels.setPixelColor(i,Color(5,5,5));
      }
    

    for (int i = 0; i<30; i++)
      {
        pixels.setPixelColor(me[i],Color(255,0,0));
      }

    pixels.show(); 
  
    lightState = LIGHT_STATE_END; 
    delay_light = millis(); 
     
  break; 

  case LIGHT_STATE_END:

    lastDrawUpdate = millis(); 
    resetLEDs(); 
    lightState = LIGHT_STATE_BAGGROUND; 

  default:

    lastDrawUpdate = millis();
    lightState = LIGHT_STATE_BAGGROUND;
  break;
  }


  
  
  
  

  
  

  
  



   
}

void attentionLight1()
{
  byte r = random(1,255);
  byte g = random(1,255);
  byte b = random(1,255); 
  byte n = random(199); 

  pixels.setPixelColor(n,r,g,b); 
  byte r1 = random(1,255);
  byte g1 = random(1,255);
  byte b1 = random(1,255); 
  byte n1 = random(199); 
  
  pixels.setPixelColor(n1,r1,g1,b1); 
  pixels.show(); 
  delay_light = millis(); 
  delay_light_high = false; 
  
  
}

bool is_waiting_for_ms(unsigned long wait_time_ms)
{
  static bool is_waiting = false; 
  static unsigned long start_ms = 0; 
  static unsigned long static_wait_time_ms = 0;
  if (false == is_waiting)
  {
    start_ms = millis(); 
    is_waiting = true; 
    static_wait_time_ms = wait_time_ms; 
  }
  else
  {
    if (millis() - static_wait_time_ms >= start_ms )
    {
      is_waiting = false; 
    }
  }
  return is_waiting; 
}